#
#(C) 2008-2011 Advanced Micro Devices, Inc. All Rights Reserved.
#  
#  This file is part of AMD LibM 3.0.
#  
#  AMD LibM 3.0 is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2.1 of the License, or (at your option) any later version.
#  
#  AMD LibM 3.0 is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#  
#  You should have received a copy of the GNU Lesser General Public
#  License along with AMD LibM 3.0.  If not, see
#  <http://www.gnu.org/licenses/>.
#  
#

#include "fn_macros.h"
#define fname FN_PROTOTYPE_BDOZR(sinf)
#define fname_special   _sinf_special@PLT
#define fname_special_underflow _sin_special_underflow@PLT

#include "trig_func.h"


.equ    r,    0x30                               # pointer to r for amd_remainder_piby2
.equ    rr,   0x40                               # pointer to rr for amd_remainder_piby2
.equ    region,    0x50                               # pointer to region for amd_remainder_piby2
.equ    store_input, 0x60
.equ    ret_addr, 0x70
.equ   stack_size, 0x88



#ifdef __ELF__
.section .note.GNU-stack,"",@progbits
#endif

.text
.align 32
.p2align 4,,15
.globl fname
.type fname,@function
fname:

    	sub     $stack_size, %rsp

sin_early_exit_s:

    MOVD  %xmm0,%eax
    MOV   %eax,%r9d
    MOV   .L__inf_mask_32(%rip),%r8d
    AND   %r8d,%eax
    CMP   %r8d,%eax
    JZ    sinf_naninf

    VCVTPS2PD    %xmm0,%xmm0
    VMOVAPD   %xmm0, store_input(%rsp)

    MOVD  %xmm0,%r9
    AND    .L__sign_mask(%rip),%r9                          #clear sign

sinf_early_exit_s_1:

    CMP   .L_mask_3fe(%rip),%r9
    JG    range_reduce
    CMP   .L_mask_3f8(%rip),%r9
    JGE    compute_sinf_pyby_4
    CMP   .L_mask_3f2(%rip),%r9
    JGE   compute_x_xxx_0_1666


##  Additional code to raise underflow exception
    xor   %rax,%rax
    cmp   %rax,%r9	
    je    return_sinf_s
    call  fname_special_underflow  
    JMP   return_sinf_s

compute_x_xxx_0_1666:
    VMULSD    %xmm0,%xmm0,%xmm1                         #
    VMULSD    %xmm1,%xmm0,%xmm1                         # %xmm1 x3
    VFNMADDSD  %xmm0,.L_point_166(%rip),%xmm1,%xmm0         #  x - x*x*x*0.166666666666666666#
    JMP return_sinf_s

compute_sinf_pyby_4:
    sin_piby4_sf
    JMP    return_sinf_s

range_reduce:
    VMOVQ   %r9,%xmm0                                # %r9 x with the sign cleared
    cmp  .L_e_5(%rip),%r9
    JGE  sinf_remainder_piby2

#sinf_range_e_5_s:
    range_e_5_sf
    JMP sinf_exit_s

sinf_remainder_piby2:
    call_remainder_piby2_f

sinf_exit_s:

    VANDPD     .L_int_one(%rip),%xmm11,%xmm3
    VPCOMEQUQ    .L_int_one(%rip),%xmm3,%xmm3

    VMOVDDUP    %xmm3,%xmm3
    VPTEST    %xmm3,%xmm3

    JZ    sin_piby4_compute
    JMP    cos_piby4_compute

sin_piby4_compute:
    sin_piby4_sf
    JMP sinf_exit_s_1

cos_piby4_compute:
    cos_piby4_sf

sinf_exit_s_1:

    VPCOMEQUQ    .L_int_two(%rip),%xmm11,%xmm2
    VPCOMEQUQ    .L_int_three(%rip),%xmm11,%xmm3
    VORPD    %xmm2,%xmm3,%xmm3
    VANDNPD    .L_signbit(%rip),%xmm3,%xmm3
    VXORPD    %xmm0,%xmm3,%xmm0

    VMOVAPD    store_input(%rsp),%xmm1
    VANDNPD    .L_signbit(%rip),%xmm1,%xmm1
    VXORPD    %xmm1,%xmm0,%xmm0

return_sinf_s:

    VCVTPD2PS    %xmm0,%xmm0
    add      $stack_size, %rsp
    ret

sinf_naninf:
    call    fname_special
    add      $stack_size, %rsp
    ret

